from dataclasses import dataclass, field
from textwrap import dedent
from typing import Literal

import pytest

import pybag
from pybag.schema.ros2msg import (
    Array,
    Complex,
    Primitive,
    Ros2MsgError,
    Ros2MsgSchemaEncoder,
    Schema,
    SchemaConstant,
    SchemaField,
    Sequence,
    String
)


def test_serialize_dataclass_primitives() -> None:
    @dataclass
    class Example:
        __msg_name__ = 'tests/msgs/Example'

        integer: pybag.int32
        name: pybag.string
        flag: pybag.bool

    obj = Example(42, "hi", True)
    schema, sub_schemas = Ros2MsgSchemaEncoder().parse_schema(obj)

    assert schema.name == 'tests/msgs/Example'
    assert len(schema.fields) == 3
    assert schema.fields['integer'] == SchemaField(Primitive('int32'), default=None)
    assert schema.fields['name'] == SchemaField(String('string'), default=None)
    assert schema.fields['flag'] == SchemaField(Primitive('bool'), default=None)
    assert len(sub_schemas) == 0  # No sub-schemas for primitive types


def test_serialize_dataclass_primitives_with_default() -> None:
    @dataclass
    class ExampleDefault:
        __msg_name__ = 'tests/msgs/ExampleDefault'

        integer: pybag.int32 = 42
        name: pybag.string = "hi"
        flag: pybag.bool = True

    obj = ExampleDefault()
    schema, sub_schemas = Ros2MsgSchemaEncoder().parse_schema(obj)
    assert schema.name == 'tests/msgs/ExampleDefault'
    assert len(schema.fields) == 3
    assert schema.fields['integer'] == SchemaField(Primitive('int32'), default=42)
    assert schema.fields['name'] == SchemaField(String('string'), default="hi")
    assert schema.fields['flag'] == SchemaField(Primitive('bool'), default=True)
    assert len(sub_schemas) == 0  # No sub-schemas for primitive types


def test_serialize_array() -> None:
    @dataclass
    class ArrayExample:
        __msg_name__ = 'tests/msgs/ArrayExample'
        numbers: pybag.Array[pybag.int32]

    obj = ArrayExample(numbers=[1, 2, 3])
    schema, sub_schemas = Ros2MsgSchemaEncoder().parse_schema(obj)

    assert schema.name == 'tests/msgs/ArrayExample'
    assert len(schema.fields) == 1
    assert schema.fields['numbers'] == SchemaField(Sequence(Primitive('int32')), default=None)
    assert len(sub_schemas) == 0  # No sub-schemas for primitive types


def test_serialize_array_with_length() -> None:
    @dataclass
    class ArrayExample:
        __msg_name__ = 'tests/msgs/ArrayExample'
        numbers: pybag.Array[pybag.int32, Literal[3]]

    obj = ArrayExample(numbers=[1, 2, 3])
    schema, sub_schemas = Ros2MsgSchemaEncoder().parse_schema(obj)

    assert schema.name == 'tests/msgs/ArrayExample'
    assert len(schema.fields) == 1
    assert schema.fields['numbers'] == SchemaField(Array(Primitive('int32'), length=3, is_bounded=False), default=None)
    assert len(sub_schemas) == 0  # No sub-schemas for primitive types


def test_serialize_array_with_default() -> None:
    @dataclass
    class ArrayExample:
        __msg_name__ = 'tests/msgs/ArrayExample'
        numbers: pybag.Array[pybag.int32] = field(default_factory=lambda: [1, 2, 3])

    obj = ArrayExample()
    schema, sub_schemas = Ros2MsgSchemaEncoder().parse_schema(obj)

    assert schema.name == 'tests/msgs/ArrayExample'
    assert len(schema.fields) == 1
    assert schema.fields['numbers'] == SchemaField(Sequence(Primitive('int32')), default=[1, 2, 3])
    assert len(sub_schemas) == 0  # No sub-schemas for primitive types


def test_serialize_python_types_int() -> None:
    @dataclass
    class Missing:
        __msg_name__ = 'tests/msgs/Missing'
        value: int

    with pytest.raises(Ros2MsgError):
        Ros2MsgSchemaEncoder().parse_schema(Missing)


def test_serialize_python_types_str() -> None:
    @dataclass
    class Unsupported:
        __msg_name__ = 'tests/msgs/Unsupported'
        value: str

    with pytest.raises(Ros2MsgError):
        Ros2MsgSchemaEncoder().parse_schema(Unsupported)


def test_serialize_constants() -> None:
    @dataclass(kw_only=True)
    class Example:
        __msg_name__ = 'tests/msgs/ExampleConst'
        FOO: pybag.Constant[pybag.int32] = 1
        bar: pybag.int32

    obj = Example(bar=42)
    schema, sub_schemas = Ros2MsgSchemaEncoder().parse_schema(obj)

    assert schema.name == 'tests/msgs/ExampleConst'
    assert schema.fields['FOO'] == SchemaConstant(Primitive('int32'), value=1)
    assert schema.fields['bar'] == SchemaField(Primitive('int32'), default=None)
    assert len(sub_schemas) == 0

    encoded = Ros2MsgSchemaEncoder().encode(obj)
    assert encoded == b'int32 FOO=1\nint32 bar\n'


def test_serialize_nested_complex_array() -> None:
    @dataclass()
    class MultiArrayDimension:
        __msg_name__ = 'std_msgs/msg/MultiArrayDimension'

        label: pybag.string
        size: pybag.uint32
        stride: pybag.uint32

    @dataclass()
    class MultiArrayLayout:
        __msg_name__ = 'std_msgs/msg/MultiArrayLayout'

        dim: pybag.Array[pybag.Complex[MultiArrayDimension]]
        data_offset: pybag.uint32

    @dataclass()
    class Float32MultiArray:
        __msg_name__ = 'std_msgs/msg/Float32MultiArray'

        layout: pybag.Complex[MultiArrayLayout]
        data: pybag.Array[pybag.float32]

    obj = Float32MultiArray(
        layout=MultiArrayLayout(
            dim=[MultiArrayDimension(label='test', size=10, stride=5)],
            data_offset=0
        ),
        data=[1.1, 2.2, 3.3]
    )
    schema, sub_schemas = Ros2MsgSchemaEncoder().parse_schema(obj)

    assert schema.name == 'std_msgs/msg/Float32MultiArray'
    assert schema.fields['layout'] == SchemaField(Complex('std_msgs/msg/MultiArrayLayout'), default=None)
    assert schema.fields['data'] == SchemaField(Sequence(Primitive('float32')), default=None)
    assert len(sub_schemas) == 2
    assert sub_schemas['std_msgs/msg/MultiArrayLayout'] == Schema(
        name='std_msgs/msg/MultiArrayLayout',
        fields={
            'dim': SchemaField(Sequence(Complex('std_msgs/msg/MultiArrayDimension')), default=None),
            'data_offset': SchemaField(Primitive('uint32'), default=None)
        }
    )
    assert sub_schemas['std_msgs/msg/MultiArrayDimension'] == Schema(
        name='std_msgs/msg/MultiArrayDimension',
        fields={
            'label': SchemaField(String('string'), default=None),
            'size': SchemaField(Primitive('uint32'), default=None),
            'stride': SchemaField(Primitive('uint32'), default=None)
        }
    )

    encoded = Ros2MsgSchemaEncoder().encode(obj)
    print(encoded)
    assert encoded == dedent("""\
        std_msgs/MultiArrayLayout layout
        float32[] data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        std_msgs/MultiArrayDimension[] dim
        uint32 data_offset
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        string label
        uint32 size
        uint32 stride
        """).encode()


def test_serialize_python_types_float() -> None:
    @dataclass
    class Unsupported:
        __msg_name__ = 'tests/msgs/Unsupported'
        value: float

    with pytest.raises(Ros2MsgError):
        Ros2MsgSchemaEncoder().parse_schema(Unsupported)


def test_serialize_python_types_bool() -> None:
    @dataclass
    class Unsupported:
        __msg_name__ = 'tests/msgs/Unsupported'
        value: bool

    with pytest.raises(Ros2MsgError):
        Ros2MsgSchemaEncoder().parse_schema(Unsupported)



"""MCAP info CLI command."""

from pathlib import Path
from textwrap import dedent

from pybag.mcap.record_reader import McapRecordReaderFactory


def _ns_to_seconds(ns: int) -> float:
    """Convert nanoseconds to seconds."""
    return ns / 1_000_000_000


def _format_duration(duration_ns: int) -> str:
    """Format duration in a human-readable way."""
    duration_s = _ns_to_seconds(duration_ns)

    if duration_s < 1:
        return f"{duration_s * 1000:.2f} ms"
    elif duration_s < 60:
        return f"{duration_s:.2f} s"
    elif duration_s < 3600:
        minutes = int(duration_s // 60)
        seconds = duration_s % 60
        return f"{minutes}m {seconds:.2f}s"
    else:
        hours = int(duration_s // 3600)
        minutes = int((duration_s % 3600) // 60)
        seconds = duration_s % 60
        return f"{hours}h {minutes}m {seconds:.2f}s"


def info_mcap(input_path: str | Path) -> None:
    """Display information about an MCAP file."""
    input_path = Path(input_path).resolve()

    with McapRecordReaderFactory.from_file(input_path) as reader:
        # Get header information
        header = reader.get_header()

        # Get statistics
        stats = reader.get_statistics()

        # Get channels and schemas
        channels = reader.get_channels()
        schemas = reader.get_schemas()

        # Calculate duration
        duration_ns = stats.message_end_time - stats.message_start_time

        # Print file information
        print(f"File: {input_path.name}")
        print(f"Path: {input_path}")
        print()

        # Print time information
        print("Time Information:")
        print(f"  Start time:  {_ns_to_seconds(stats.message_start_time):.6f} s")
        print(f"  End time:    {_ns_to_seconds(stats.message_end_time):.6f} s")
        print(f"  Duration:    {_format_duration(duration_ns)}")
        print()

        # Print general statistics
        print("Statistics:")
        print(f"  Profile:        {header.profile}")
        print(f"  Library:        {header.library}")
        print(f"  Messages:       {stats.message_count:,}")
        print(f"  Channels:       {stats.channel_count}")
        print(f"  Schemas:        {stats.schema_count}")
        print(f"  Chunks:         {stats.chunk_count}")
        print(f"  Attachments:    {stats.attachment_count}")
        print(f"  Metadata:       {stats.metadata_count}")
        print()

        # Print topic information
        if channels:
            print("Topics:")

            # Build a list of topic info
            topic_info = []
            for channel_id, channel in channels.items():
                msg_count = stats.channel_message_counts.get(channel_id, 0)

                # Calculate frequency (messages per second)
                frequency = msg_count / _ns_to_seconds(duration_ns) if duration_ns > 0 else 0

                # Get schema encoding if available
                schema_encoding = "N/A"
                if channel.schema_id != 0 and channel.schema_id in schemas:
                    schema = schemas[channel.schema_id]
                    schema_encoding = schema.encoding

                topic_info.append({
                    'topic': channel.topic,
                    'messages': msg_count,
                    'encoding': channel.message_encoding,
                    'schema_encoding': schema_encoding,
                    'frequency': frequency,
                    'schema_name': schemas[channel.schema_id].name if channel.schema_id != 0 and channel.schema_id in schemas else "N/A"
                })

            # Sort by topic name
            topic_info.sort(key=lambda x: x['topic'])

            # Find max widths for alignment
            max_topic_len = max(len(t['topic']) for t in topic_info)
            max_msgs_len = max(len(f"{t['messages']:,}") for t in topic_info)

            # Print header
            print(f"  {'Topic':<{max_topic_len}}  {'Messages':>{max_msgs_len}}  {'Freq (Hz)':>10}  {'Encoding':<15}  Schema")
            print(f"  {'-' * max_topic_len}  {'-' * max_msgs_len}  {'-' * 10}  {'-' * 15}  {'-' * 20}")

            # Print each topic
            for info in topic_info:
                topic = info['topic']
                messages = f"{info['messages']:,}"
                freq = f"{info['frequency']:.2f}" if info['frequency'] < 1000 else f"{info['frequency']:.1f}"
                encoding = info['encoding']
                schema = info['schema_name']

                print(f"  {topic:<{max_topic_len}}  {messages:>{max_msgs_len}}  {freq:>10}  {encoding:<15}  {schema}")


def add_parser(subparsers) -> None:
    parser = subparsers.add_parser(
        "info",
        help="Display information about an MCAP file.",
        description=dedent("""
            Display statistics and information about an MCAP file including:
            - File name and path
            - Start/end time and duration
            - Profile and library
            - Message, channel, and schema counts
            - Per-topic statistics (message count, encoding, frequency)
        """)
    )
    parser.add_argument("input", help="Path to MCAP file (*.mcap)")
    parser.set_defaults(func=lambda args: info_mcap(args.input))



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pybag.types import Message

from pybag.mcap.records import SchemaRecord

# Map ROS2 types to Python types
# https://docs.ros.org/en/foxy/Concepts/About-ROS-Interfaces.html#field-types
PRIMITIVE_TYPE_MAP = {
    'bool': bool,
    'byte': bytes,
    'char': str,
    'float32': float,
    'float64': float,
    'int8': int,
    'uint8': int,
    'int16': int,
    'uint16': int,
    'int32': int,
    'uint32': int,
    'int64': int,
    'uint64': int,
}
STRING_TYPE_MAP = {
    'string': str,
    'wstring': str,
}


@dataclass
class SchemaFieldType(ABC):
    ...


@dataclass
class Primitive(SchemaFieldType):
    type: str

    @classmethod
    def is_primitive(cls, type: str) -> bool:
        return type in PRIMITIVE_TYPE_MAP


@dataclass
class String(SchemaFieldType):
    type: str = 'string'
    max_length: int | None = None


@dataclass
class Array(SchemaFieldType):
    type: 'SchemaFieldType'
    length: int
    is_bounded: bool = False  # False == fixed length


@dataclass
class Sequence(SchemaFieldType):
    type: 'SchemaFieldType'


@dataclass
class Complex(SchemaFieldType):
    type: str


@dataclass
class SchemaEntry(ABC):
    ...


@dataclass
class SchemaConstant(SchemaEntry):
    type: SchemaFieldType
    value: int | float | bool | str | bytes


@dataclass
class SchemaField(SchemaEntry):
    type: SchemaFieldType
    default: Any = None


@dataclass
class Schema:
    name: str
    fields: dict[str, SchemaEntry]


class SchemaDecoder(ABC):
    @abstractmethod
    def parse_schema(self, schema: SchemaRecord) -> tuple[Schema, dict[str, Schema]]:
        """Decode a schema into a Python object."""
        ...  # pragma: no cover


class SchemaEncoder(ABC):
    @classmethod
    @abstractmethod
    def encoding(cls) -> str:
        """The encoding to use for the schema."""
        ...  # pragma: no cover

    @abstractmethod
    def encode(self, schema: Message | type[Message]) -> bytes:
        """Encode a schema into a bytes object."""
        ...  # pragma: no cover

    @abstractmethod
    def parse_schema(self, schema: Message | type[Message]) -> tuple[Schema, dict[str, Schema]]:
        """Parse a schema into a Python object."""
        ...  # pragma: no cover



from typing import Callable

from pybag.encoding import MessageDecoder
from pybag.encoding.cdr import CdrDecoder
from pybag.mcap.records import ChannelRecord, MessageRecord, SchemaRecord
from pybag.schema import SchemaDecoder
from pybag.schema.compiler import compile_schema
from pybag.schema.ros2msg import Ros2MsgSchemaDecoder


class MessageDeserializer:
    """Abstract base class for message deserializers."""

    def __init__(
        self,
        schema_decoder: SchemaDecoder,
        message_decoder: type[MessageDecoder],
    ):
        self._schema_decoder = schema_decoder
        self._message_decoder = message_decoder
        self._compiled: dict[int, Callable[[MessageDecoder], type]] = {}

    def deserialize_message(self, message: MessageRecord, schema: SchemaRecord) -> type:
        """Deserialize a message using the provided schema."""
        decoder = self._message_decoder(message.data)
        if schema.id not in self._compiled:
            msg_schema, schema_msgs = self._schema_decoder.parse_schema(schema)
            self._compiled[schema.id] = compile_schema(msg_schema, schema_msgs)
        return self._compiled[schema.id](decoder)


class MessageDeserializerFactory:
    """Factory for creating message deserializers."""

    @staticmethod
    def from_profile(profile: str) -> MessageDeserializer | None:
        if profile == "ros2":
            return MessageDeserializer(Ros2MsgSchemaDecoder(), CdrDecoder)
        return None

    @staticmethod
    def from_channel(channel: ChannelRecord, schema: SchemaRecord) -> MessageDeserializer | None:
        if channel.message_encoding == "cdr" and schema.encoding == "ros2msg":
            return MessageDeserializer(Ros2MsgSchemaDecoder(), CdrDecoder)
        return None



import zlib

from pybag.io.raw_reader import BaseReader

DEFAULT_CRC_CHUNK_SIZE = 8 * 1024 * 1024  # 8 MiB

class McapInvalidCrcError(Exception):
    """Exception raised when a CRC is invalid."""


def validate_crc(data: bytes, crc: int) -> bool:
    """Validate CRC values of an MCAP file.
    """
    return zlib.crc32(data) == crc


def assert_crc(data: bytes, crc: int) -> None:
    """Assert CRC values of an MCAP file.
    """
    if not validate_crc(data, crc):
        raise McapInvalidCrcError(f'Invalid CRC for data')


def compute_crc_batched(reader: BaseReader, num_bytes: int, chunk_size: int = DEFAULT_CRC_CHUNK_SIZE) -> int:
    """Compute CRC32 over the next num_bytes from current reader position in chunks."""
    remaining = num_bytes
    crc_value = 0
    while remaining > 0:
        read_size = min(remaining, chunk_size)
        if not (chunk := reader.read(read_size)):
            break
        crc_value = zlib.crc32(chunk, crc_value)
        remaining -= read_size
    return crc_value



from pathlib import Path

import pybag.ros2.humble.std_msgs as std_msgs
from pybag.cli.main import main
from pybag.mcap_reader import McapFileReader
from pybag.mcap_writer import McapFileWriter


def test_merge_cli(tmp_path: Path) -> None:
    input1 = tmp_path / "one.mcap"
    input2 = tmp_path / "two.mcap"
    output = tmp_path / "merged.mcap"

    with McapFileWriter.open(input1) as writer:
        writer.write_message("/one", 1, std_msgs.String(data="a"))
    with McapFileWriter.open(input2) as writer:
        writer.write_message("/two", 2, std_msgs.String(data="b"))

    main(["merge", str(input1), str(input2), "-o", str(output)])

    with McapFileReader.from_file(output) as reader:
        assert set(reader.get_topics()) == set(["/one", "/two"])
        # Get /one topic message
        messages = list(reader.messages("/one"))
        assert len(messages) == 1
        # Get /two topic message
        messages = list(reader.messages("/two"))
        assert len(messages) == 1


def test_merge_cli_same_topic(tmp_path: Path) -> None:
    input1 = tmp_path / "one.mcap"
    input2 = tmp_path / "two.mcap"
    output = tmp_path / "merged.mcap"

    with McapFileWriter.open(input1) as writer:
        writer.write_message("/one", 1, std_msgs.String(data="a"))
    with McapFileWriter.open(input2) as writer:
        writer.write_message("/one", 2, std_msgs.String(data="b"))

    main(["merge", str(input1), str(input2), "-o", str(output)])

    with McapFileReader.from_file(output) as reader:
        assert set(reader.get_topics()) == set(["/one"])
        # Get /one topic message
        messages = list(reader.messages("/one"))
        assert len(messages) == 2



"""Test reading geometry_msgs messages written with the official MCAP writer."""

from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from typing import Any

from mcap_ros2.writer import Writer as McapWriter

from pybag.mcap_reader import McapFileReader


def _write_mcap(temp_dir: str, msg: Any, msgtype: str, schema_text: str) -> Path:
    path = Path(temp_dir) / "test.mcap"
    with open(path, "wb") as f:
        writer = McapWriter(f)
        schema = writer.register_msgdef(msgtype, schema_text)
        writer.write_message(
            topic="/rosbags",
            schema=schema,
            message=msg,
            log_time=0,
            publish_time=0,
            sequence=0,
        )
        writer.finish()
    return path


def test_geometry_msgs_accel():
    msgtype = "geometry_msgs/Accel"
    schema = dedent("""
        # This expresses acceleration in free space broken into its linear and angular parts.
        Vector3 linear
        Vector3 angular
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "linear": {"x": 1.0, "y": 2.0, "z": 3.0},
            "angular": {"x": 4.0, "y": 5.0, "z": 6.0},
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.linear.x == 1.0
    assert messages[0].data.linear.y == 2.0
    assert messages[0].data.linear.z == 3.0
    assert messages[0].data.angular.x == 4.0
    assert messages[0].data.angular.y == 5.0
    assert messages[0].data.angular.z == 6.0


def test_geometry_msgs_accel_stamped():
    msgtype = "geometry_msgs/AccelStamped"
    schema = dedent("""
        # An accel with reference coordinate frame and timestamp
        std_msgs/Header header
        Accel accel
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Accel
        # This expresses acceleration in free space broken into its linear and angular parts.
        Vector3 linear
        Vector3 angular
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "accel": {
                "linear": {"x": 1.0, "y": 2.0, "z": 3.0},
                "angular": {"x": 4.0, "y": 5.0, "z": 6.0}
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.accel.linear.x == 1.0
    assert messages[0].data.accel.linear.y == 2.0
    assert messages[0].data.accel.linear.z == 3.0
    assert messages[0].data.accel.angular.x == 4.0
    assert messages[0].data.accel.angular.y == 5.0
    assert messages[0].data.accel.angular.z == 6.0


def test_geometry_msgs_accel_with_covariance():
    msgtype = "geometry_msgs/AccelWithCovariance"
    schema = dedent("""
        # This expresses acceleration in free space with uncertainty.

        Accel accel

        # Row-major representation of the 6x6 covariance matrix
        # The orientation parameters use a fixed-axis representation.
        # In order, the parameters are:
        # (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
        float64[36] covariance
        ================================================================================
        MSG: geometry_msgs/Accel
        # This expresses acceleration in free space broken into its linear and angular parts.
        Vector3 linear
        Vector3 angular
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "accel": {
                "linear": {"x": 1.0, "y": 2.0, "z": 3.0},
                "angular": {"x": 4.0, "y": 5.0, "z": 6.0}
            },
            "covariance": [0.0] * 36
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.accel.linear.x == 1.0
    assert messages[0].data.accel.linear.y == 2.0
    assert messages[0].data.accel.linear.z == 3.0
    assert messages[0].data.accel.angular.x == 4.0
    assert messages[0].data.accel.angular.y == 5.0
    assert messages[0].data.accel.angular.z == 6.0
    assert messages[0].data.covariance == [0.0] * 36


def test_geometry_msgs_accel_with_covariance_stamped():
    msgtype = "geometry_msgs/AccelWithCovarianceStamped"
    schema = dedent("""
        # An accel with covariance with reference coordinate frame and timestamp
        std_msgs/Header header
        AccelWithCovariance accel
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/AccelWithCovariance
        # This expresses acceleration in free space with uncertainty.

        Accel accel

        # Row-major representation of the 6x6 covariance matrix
        # The orientation parameters use a fixed-axis representation.
        # In order, the parameters are:
        # (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
        float64[36] covariance
        ================================================================================
        MSG: geometry_msgs/Accel
        # This expresses acceleration in free space broken into its linear and angular parts.
        Vector3 linear
        Vector3 angular
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "accel": {
                "accel": {
                    "linear": {"x": 1.0, "y": 2.0, "z": 3.0},
                    "angular": {"x": 4.0, "y": 5.0, "z": 6.0}
                },
                "covariance": [0.0] * 36
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.accel.accel.linear.x == 1.0
    assert messages[0].data.accel.accel.linear.y == 2.0
    assert messages[0].data.accel.accel.linear.z == 3.0
    assert messages[0].data.accel.accel.angular.x == 4.0
    assert messages[0].data.accel.accel.angular.y == 5.0
    assert messages[0].data.accel.accel.angular.z == 6.0
    assert messages[0].data.accel.covariance == [0.0] * 36


def test_geometry_msgs_inertia():
    msgtype = "geometry_msgs/Inertia"
    schema = dedent("""
        # Mass [kg]
        float64 m

        # Center of mass [m]
        geometry_msgs/Vector3 com

        # Inertia Tensor [kg-m^2]
        #     | ixx ixy ixz |
        # I = | ixy iyy iyz |
        #     | ixz iyz izz |
        float64 ixx
        float64 ixy
        float64 ixz
        float64 iyy
        float64 iyz
        float64 izz
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "m": 10.5,
            "com": {"x": 1.0, "y": 2.0, "z": 3.0},
            "ixx": 1.1,
            "ixy": 1.2,
            "ixz": 1.3,
            "iyy": 2.2,
            "iyz": 2.3,
            "izz": 3.3
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.m == 10.5
    assert messages[0].data.com.x == 1.0
    assert messages[0].data.com.y == 2.0
    assert messages[0].data.com.z == 3.0
    assert messages[0].data.ixx == 1.1
    assert messages[0].data.ixy == 1.2
    assert messages[0].data.ixz == 1.3
    assert messages[0].data.iyy == 2.2
    assert messages[0].data.iyz == 2.3
    assert messages[0].data.izz == 3.3


def test_geometry_msgs_inertia_stamped():
    msgtype = "geometry_msgs/InertiaStamped"
    schema = dedent("""
        # An inertia with reference coordinate frame and timestamp
        std_msgs/Header header
        Inertia inertia
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Inertia
        # Mass [kg]
        float64 m

        # Center of mass [m]
        geometry_msgs/Vector3 com

        # Inertia Tensor [kg-m^2]
        #     | ixx ixy ixz |
        # I = | ixy iyy iyz |
        #     | ixz iyz izz |
        float64 ixx
        float64 ixy
        float64 ixz
        float64 iyy
        float64 iyz
        float64 izz
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "inertia": {
                "m": 10.5,
                "com": {"x": 1.0, "y": 2.0, "z": 3.0},
                "ixx": 1.1,
                "ixy": 1.2,
                "ixz": 1.3,
                "iyy": 2.2,
                "iyz": 2.3,
                "izz": 3.3
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.inertia.m == 10.5
    assert messages[0].data.inertia.com.x == 1.0
    assert messages[0].data.inertia.com.y == 2.0
    assert messages[0].data.inertia.com.z == 3.0
    assert messages[0].data.inertia.ixx == 1.1
    assert messages[0].data.inertia.ixy == 1.2
    assert messages[0].data.inertia.ixz == 1.3
    assert messages[0].data.inertia.iyy == 2.2
    assert messages[0].data.inertia.iyz == 2.3
    assert messages[0].data.inertia.izz == 3.3


def test_geometry_msgs_point():
    msgtype = "geometry_msgs/Point"
    schema = dedent("""
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"x": 1.0, "y": 2.0, "z": 3.0}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.x == 1.0
    assert messages[0].data.y == 2.0
    assert messages[0].data.z == 3.0


def test_geometry_msgs_point32():
    msgtype = "geometry_msgs/Point32"
    schema = dedent("""
        # This contains the position of a point in free space(with 32 bits of precision).
        # It is recommended to use Point wherever possible instead of Point32.
        #
        # This recommendation is to promote interoperability.
        #
        # This message is designed to take up less space when sending
        # lots of points at once, as in the case of a PointCloud.

        float32 x
        float32 y
        float32 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"x": 1.0, "y": 2.0, "z": 3.0}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.x == 1.0
    assert messages[0].data.y == 2.0
    assert messages[0].data.z == 3.0


def test_geometry_msgs_point_stamped():
    msgtype = "geometry_msgs/PointStamped"
    schema = dedent("""
        # A Point with reference coordinate frame and timestamp
        std_msgs/Header header
        Point point
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "point": {"x": 1.0, "y": 2.0, "z": 3.0}
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.point.x == 1.0
    assert messages[0].data.point.y == 2.0
    assert messages[0].data.point.z == 3.0


def test_geometry_msgs_polygon():
    msgtype = "geometry_msgs/Polygon"
    schema = dedent("""
        # A specification of a polygon where the first and last points are assumed to be connected
        Point32[] points
        ================================================================================
        MSG: geometry_msgs/Point32
        # This contains the position of a point in free space(with 32 bits of precision).
        # It is recommended to use Point wherever possible instead of Point32.
        #
        # This recommendation is to promote interoperability.
        #
        # This message is designed to take up less space when sending
        # lots of points at once, as in the case of a PointCloud.

        float32 x
        float32 y
        float32 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "points": [
                {"x": 1.0, "y": 2.0, "z": 3.0},
                {"x": 4.0, "y": 5.0, "z": 6.0}
            ]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.points) == 2
    assert messages[0].data.points[0].x == 1.0
    assert messages[0].data.points[0].y == 2.0
    assert messages[0].data.points[0].z == 3.0
    assert messages[0].data.points[1].x == 4.0
    assert messages[0].data.points[1].y == 5.0
    assert messages[0].data.points[1].z == 6.0


def test_geometry_msgs_polygon_instance():
    msgtype = "geometry_msgs/PolygonInstance"
    schema = dedent("""
        # A Polygon with an ID for discrimination between multiple polygons
        geometry_msgs/Polygon polygon
        int64 id
        ================================================================================
        MSG: geometry_msgs/Polygon
        # A specification of a polygon where the first and last points are assumed to be connected
        Point32[] points
        ================================================================================
        MSG: geometry_msgs/Point32
        # This contains the position of a point in free space(with 32 bits of precision).
        # It is recommended to use Point wherever possible instead of Point32.
        #
        # This recommendation is to promote interoperability.
        #
        # This message is designed to take up less space when sending
        # lots of points at once, as in the case of a PointCloud.

        float32 x
        float32 y
        float32 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "polygon": {"points": [{"x": 1.0, "y": 2.0, "z": 3.0}]},
            "id": 42
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.id == 42
    assert len(messages[0].data.polygon.points) == 1
    assert messages[0].data.polygon.points[0].x == 1.0
    assert messages[0].data.polygon.points[0].y == 2.0
    assert messages[0].data.polygon.points[0].z == 3.0


def test_geometry_msgs_polygon_instance_stamped():
    msgtype = "geometry_msgs/PolygonInstanceStamped"
    schema = dedent("""
        std_msgs/Header header
        geometry_msgs/PolygonInstance polygon
        ================================================================================
        MSG: geometry_msgs/PolygonInstance
        geometry_msgs/Polygon polygon
        int64 id
        ================================================================================
        MSG: geometry_msgs/Polygon
        Point32[] points
        ================================================================================
        MSG: geometry_msgs/Point32
        # This contains the position of a point in free space(with 32 bits of precision).
        # It is recommended to use Point wherever possible instead of Point32.
        #
        # This recommendation is to promote interoperability.
        #
        # This message is designed to take up less space when sending
        # lots of points at once, as in the case of a PointCloud.

        float32 x
        float32 y
        float32 z
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
    """)
    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "polygon": {
                "polygon": {"points": [{"x": 1.0, "y": 2.0, "z": 3.0}]},
                "id": 42
            },
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.polygon.id == 42
    assert len(messages[0].data.polygon.polygon.points) == 1
    assert messages[0].data.polygon.polygon.points[0].x == 1.0
    assert messages[0].data.polygon.polygon.points[0].y == 2.0
    assert messages[0].data.polygon.polygon.points[0].z == 3.0


def test_geometry_msgs_polygon_stamped():
    msgtype = "geometry_msgs/PolygonStamped"
    schema = dedent("""
        std_msgs/Header header
        Polygon polygon
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Polygon
        Point32[] points
        ================================================================================
        MSG: geometry_msgs/Point32
        # This contains the position of a point in free space(with 32 bits of precision).
        # It is recommended to use Point wherever possible instead of Point32.
        #
        # This recommendation is to promote interoperability.
        #
        # This message is designed to take up less space when sending
        # lots of points at once, as in the case of a PointCloud.

        float32 x
        float32 y
        float32 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "polygon": {
                "points": [
                    {"x": 1.0, "y": 2.0, "z": 3.0},
                    {"x": 4.0, "y": 5.0, "z": 6.0}
                ]
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert len(messages[0].data.polygon.points) == 2
    assert messages[0].data.polygon.points[0].x == 1.0
    assert messages[0].data.polygon.points[0].y == 2.0
    assert messages[0].data.polygon.points[0].z == 3.0
    assert messages[0].data.polygon.points[1].x == 4.0
    assert messages[0].data.polygon.points[1].y == 5.0
    assert messages[0].data.polygon.points[1].z == 6.0


def test_geometry_msgs_pose():
    msgtype = "geometry_msgs/Pose"
    schema = dedent("""
        # A representation of pose in free space, composed of position and orientation.

        Point position
        Quaternion orientation
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.

        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "position": {"x": 1.0, "y": 2.0, "z": 3.0},
            "orientation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.position.x == 1.0
    assert messages[0].data.position.y == 2.0
    assert messages[0].data.position.z == 3.0
    assert messages[0].data.orientation.x == 0.0
    assert messages[0].data.orientation.y == 0.0
    assert messages[0].data.orientation.z == 0.0
    assert messages[0].data.orientation.w == 1.0


def test_geometry_msgs_pose2d():
    msgtype = "geometry_msgs/Pose2D"
    schema = dedent("""
        # Deprecated as of Foxy and will potentially be removed in any following release.
        # Please use the full 3D pose.

        # In general our recommendation is to use a full 3D representation of everything and for 2D specific applications make the appropriate projections into the plane for their calculations but optimally will preserve the 3D information during processing.

        # If we have parallel copies of 2D datatypes every UI and other pipeline will end up needing to have dual interfaces to plot everything. And you will end up with not being able to use 3D tools for 2D use cases even if they're completely valid, as you'd have to reimplement it with different inputs and outputs. It's not particularly hard to plot the 2D pose or compute the yaw error for the Pose message and there are already tools and libraries that can do this for you.

        # This expresses a position and orientation on a 2D manifold.
        float64 x
        float64 y
        float64 theta
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"x": 1.0, "y": 2.0, "theta": 1.5708}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.x == 1.0
    assert messages[0].data.y == 2.0
    assert messages[0].data.theta == 1.5708


def test_geometry_msgs_pose_array():
    msgtype = "geometry_msgs/PoseArray"
    schema = dedent("""
        std_msgs/Header header
        Pose[] poses
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Pose
        # A representation of pose in free space, composed of position and orientation.

        Point position
        Quaternion orientation
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.

        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "poses": [
                {
                    "position": {"x": 1.0, "y": 2.0, "z": 3.0},
                    "orientation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
                }
            ]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert len(messages[0].data.poses) == 1
    assert messages[0].data.poses[0].position.x == 1.0
    assert messages[0].data.poses[0].position.y == 2.0
    assert messages[0].data.poses[0].position.z == 3.0
    assert messages[0].data.poses[0].orientation.x == 0.0
    assert messages[0].data.poses[0].orientation.y == 0.0
    assert messages[0].data.poses[0].orientation.z == 0.0
    assert messages[0].data.poses[0].orientation.w == 1.0


def test_geometry_msgs_pose_stamped():
    msgtype = "geometry_msgs/PoseStamped"
    schema = dedent("""
        # A Pose with reference coordinate frame and timestamp
        std_msgs/Header header
        Pose pose
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Pose
        # A representation of pose in free space, composed of position and orientation.
        Point position
        Quaternion orientation
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.
        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "pose": {
                "position": {"x": 1.0, "y": 2.0, "z": 3.0},
                "orientation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.pose.position.x == 1.0
    assert messages[0].data.pose.position.y == 2.0
    assert messages[0].data.pose.position.z == 3.0
    assert messages[0].data.pose.orientation.x == 0.0
    assert messages[0].data.pose.orientation.y == 0.0
    assert messages[0].data.pose.orientation.z == 0.0
    assert messages[0].data.pose.orientation.w == 1.0


def test_geometry_msgs_pose_with_covariance():
    msgtype = "geometry_msgs/PoseWithCovariance"
    schema = dedent("""
        # This represents a pose in free space with uncertainty.
        Pose pose

        # Row-major representation of the 6x6 covariance matrix
        # The orientation parameters use a fixed-axis representation.
        # In order, the parameters are:
        # (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
        float64[36] covariance
        ================================================================================
        MSG: geometry_msgs/Pose
        # A representation of pose in free space, composed of position and orientation.
        Point position
        Quaternion orientation
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.
        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "pose": {
                "position": {"x": 1.0, "y": 2.0, "z": 3.0},
                "orientation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
            },
            "covariance": [0.0] * 36
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.pose.position.x == 1.0
    assert messages[0].data.pose.position.y == 2.0
    assert messages[0].data.pose.position.z == 3.0
    assert messages[0].data.pose.orientation.x == 0.0
    assert messages[0].data.pose.orientation.y == 0.0
    assert messages[0].data.pose.orientation.z == 0.0
    assert messages[0].data.pose.orientation.w == 1.0
    assert messages[0].data.covariance == [0.0] * 36


def test_geometry_msgs_pose_with_covariance_stamped():
    msgtype = "geometry_msgs/PoseWithCovarianceStamped"
    schema = dedent("""
        std_msgs/Header header
        PoseWithCovariance pose
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/PoseWithCovariance
        Pose pose
        float64[36] covariance
        ================================================================================
        MSG: geometry_msgs/Pose
        # A representation of pose in free space, composed of position and orientation.

        Point position
        Quaternion orientation
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.

        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "pose": {
                "pose": {
                    "position": {"x": 1.0, "y": 2.0, "z": 3.0},
                    "orientation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
                },
                "covariance": [0.0] * 36
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.pose.pose.position.x == 1.0
    assert messages[0].data.pose.pose.position.y == 2.0
    assert messages[0].data.pose.pose.position.z == 3.0
    assert messages[0].data.pose.pose.orientation.x == 0.0
    assert messages[0].data.pose.pose.orientation.y == 0.0
    assert messages[0].data.pose.pose.orientation.z == 0.0
    assert messages[0].data.pose.pose.orientation.w == 1.0
    assert messages[0].data.pose.covariance == [0.0] * 36


def test_geometry_msgs_quaternion():
    msgtype = "geometry_msgs/Quaternion"
    schema = dedent("""
        # This represents an orientation in free space in quaternion form.
        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.x == 0.0
    assert messages[0].data.y == 0.0
    assert messages[0].data.z == 0.0
    assert messages[0].data.w == 1.0


def test_geometry_msgs_quaternion_stamped():
    msgtype = "geometry_msgs/QuaternionStamped"
    schema = dedent("""
        std_msgs/Header header
        Quaternion quaternion
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.

        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "quaternion": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.quaternion.x == 0.0
    assert messages[0].data.quaternion.y == 0.0
    assert messages[0].data.quaternion.z == 0.0
    assert messages[0].data.quaternion.w == 1.0


def test_geometry_msgs_transform():
    msgtype = "geometry_msgs/Transform"
    schema = dedent("""
        # This represents the transform between two coordinate frames in free space.
        Vector3 translation
        Quaternion rotation
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.
        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.
        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "translation": {"x": 1.0, "y": 2.0, "z": 3.0},
            "rotation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.translation.x == 1.0
    assert messages[0].data.translation.y == 2.0
    assert messages[0].data.translation.z == 3.0
    assert messages[0].data.rotation.x == 0.0
    assert messages[0].data.rotation.y == 0.0
    assert messages[0].data.rotation.z == 0.0
    assert messages[0].data.rotation.w == 1.0


def test_geometry_msgs_transform_stamped():
    msgtype = "geometry_msgs/TransformStamped"
    schema = dedent("""
        std_msgs/Header header
        string child_frame_id
        Transform transform
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Transform
        Vector3 translation
        Quaternion rotation
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.

        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "child_frame_id": "child_frame",
            "transform": {
                "translation": {"x": 1.0, "y": 2.0, "z": 3.0},
                "rotation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.child_frame_id == "child_frame"
    assert messages[0].data.transform.translation.x == 1.0
    assert messages[0].data.transform.translation.y == 2.0
    assert messages[0].data.transform.translation.z == 3.0
    assert messages[0].data.transform.rotation.x == 0.0
    assert messages[0].data.transform.rotation.y == 0.0
    assert messages[0].data.transform.rotation.z == 0.0
    assert messages[0].data.transform.rotation.w == 1.0


def test_geometry_msgs_twist():
    msgtype = "geometry_msgs/Twist"
    schema = dedent("""
        # This expresses velocity in free space broken into its linear and angular parts.

        Vector3 linear
        Vector3 angular
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "linear": {"x": 1.0, "y": 2.0, "z": 3.0},
            "angular": {"x": 4.0, "y": 5.0, "z": 6.0}
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.linear.x == 1.0
    assert messages[0].data.linear.y == 2.0
    assert messages[0].data.linear.z == 3.0
    assert messages[0].data.angular.x == 4.0
    assert messages[0].data.angular.y == 5.0
    assert messages[0].data.angular.z == 6.0


def test_geometry_msgs_twist_stamped():
    msgtype = "geometry_msgs/TwistStamped"
    schema = dedent("""
        std_msgs/Header header
        Twist twist
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Twist
        # This expresses velocity in free space broken into its linear and angular parts.

        Vector3 linear
        Vector3 angular
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "twist": {
                "linear": {"x": 1.0, "y": 2.0, "z": 3.0},
                "angular": {"x": 4.0, "y": 5.0, "z": 6.0}
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.twist.linear.x == 1.0
    assert messages[0].data.twist.linear.y == 2.0
    assert messages[0].data.twist.linear.z == 3.0
    assert messages[0].data.twist.angular.x == 4.0
    assert messages[0].data.twist.angular.y == 5.0
    assert messages[0].data.twist.angular.z == 6.0


def test_geometry_msgs_twist_with_covariance():
    msgtype = "geometry_msgs/TwistWithCovariance"
    schema = dedent("""
        # This expresses velocity in free space with uncertainty.

        Twist twist

        # Row-major representation of the 6x6 covariance matrix
        # The orientation parameters use a fixed-axis representation.
        # In order, the parameters are:
        # (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
        float64[36] covariance
        ================================================================================
        MSG: geometry_msgs/Twist
        # This expresses velocity in free space broken into its linear and angular parts.

        Vector3 linear
        Vector3 angular
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "twist": {
                "linear": {"x": 1.0, "y": 2.0, "z": 3.0},
                "angular": {"x": 4.0, "y": 5.0, "z": 6.0}
            },
            "covariance": [0.0] * 36
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.twist.linear.x == 1.0
    assert messages[0].data.twist.linear.y == 2.0
    assert messages[0].data.twist.linear.z == 3.0
    assert messages[0].data.twist.angular.x == 4.0
    assert messages[0].data.twist.angular.y == 5.0
    assert messages[0].data.twist.angular.z == 6.0
    assert messages[0].data.covariance == [0.0] * 36


def test_geometry_msgs_twist_with_covariance_stamped():
    msgtype = "geometry_msgs/TwistWithCovarianceStamped"
    schema = dedent("""
        std_msgs/Header header
        TwistWithCovariance twist
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/TwistWithCovariance
        Twist twist
        float64[36] covariance
        ================================================================================
        MSG: geometry_msgs/Twist
        # This expresses velocity in free space broken into its linear and angular parts.

        Vector3 linear
        Vector3 angular
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "twist": {
                "twist": {
                    "linear": {"x": 1.0, "y": 2.0, "z": 3.0},
                    "angular": {"x": 4.0, "y": 5.0, "z": 6.0}
                },
                "covariance": [0.0] * 36
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.twist.twist.linear.x == 1.0
    assert messages[0].data.twist.twist.linear.y == 2.0
    assert messages[0].data.twist.twist.linear.z == 3.0
    assert messages[0].data.twist.twist.angular.x == 4.0
    assert messages[0].data.twist.twist.angular.y == 5.0
    assert messages[0].data.twist.twist.angular.z == 6.0
    assert len(messages[0].data.twist.covariance) == 36


def test_geometry_msgs_vector3():
    msgtype = "geometry_msgs/Vector3"
    schema = dedent("""
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"x": 1.0, "y": 2.0, "z": 3.0}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.x == 1.0
    assert messages[0].data.y == 2.0
    assert messages[0].data.z == 3.0


def test_geometry_msgs_vector3_stamped():
    msgtype = "geometry_msgs/Vector3Stamped"
    schema = dedent("""
        std_msgs/Header header
        Vector3 vector
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "vector": {"x": 1.0, "y": 2.0, "z": 3.0}
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.vector.x == 1.0
    assert messages[0].data.vector.y == 2.0
    assert messages[0].data.vector.z == 3.0


def test_geometry_msgs_velocity_stamped():
    msgtype = "geometry_msgs/VelocityStamped"
    schema = dedent("""
        std_msgs/Header header
        string body_frame_id
        string reference_frame_id
        Twist velocity
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Twist
        # This expresses velocity in free space broken into its linear and angular parts.

        Vector3 linear
        Vector3 angular
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)
    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "velocity": {
                "linear": {"x": 1.0, "y": 2.0, "z": 3.0},
                "angular": {"x": 4.0, "y": 5.0, "z": 6.0}
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))
    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.velocity.linear.x == 1.0
    assert messages[0].data.velocity.linear.y == 2.0
    assert messages[0].data.velocity.linear.z == 3.0
    assert messages[0].data.velocity.angular.x == 4.0
    assert messages[0].data.velocity.angular.y == 5.0
    assert messages[0].data.velocity.angular.z == 6.0


def test_geometry_msgs_wrench():
    msgtype = "geometry_msgs/Wrench"
    schema = dedent("""
        # This represents force in free space, separated into its linear and angular parts.

        Vector3 force
        Vector3 torque
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "force": {"x": 1.0, "y": 2.0, "z": 3.0},
            "torque": {"x": 4.0, "y": 5.0, "z": 6.0}
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.force.x == 1.0
    assert messages[0].data.force.y == 2.0
    assert messages[0].data.force.z == 3.0
    assert messages[0].data.torque.x == 4.0
    assert messages[0].data.torque.y == 5.0
    assert messages[0].data.torque.z == 6.0


def test_geometry_msgs_wrench_stamped():
    msgtype = "geometry_msgs/WrenchStamped"
    schema = dedent("""
        std_msgs/Header header
        Wrench wrench
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Wrench
        Vector3 force
        Vector3 torque
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.

        # This is semantically different than a point.
        # A vector is always anchored at the origin.
        # When a transform is applied to a vector, only the rotational component is applied.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "test_frame"
            },
            "wrench": {
                "force": {"x": 1.0, "y": 2.0, "z": 3.0},
                "torque": {"x": 4.0, "y": 5.0, "z": 6.0}
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "test_frame"
    assert messages[0].data.wrench.force.x == 1.0
    assert messages[0].data.wrench.force.y == 2.0
    assert messages[0].data.wrench.force.z == 3.0
    assert messages[0].data.wrench.torque.x == 4.0
    assert messages[0].data.wrench.torque.y == 5.0
    assert messages[0].data.wrench.torque.z == 6.0



"""Test the reading of geometry_msgs messages."""
from array import array
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pytest
from rosbags.rosbag2 import StoragePlugin, Writer
from rosbags.typesys import Stores, get_typestore
from rosbags.typesys.store import Typestore

from pybag.mcap_reader import McapFileReader


def _find_mcap_file(temp_dir: str) -> Path:
    return next(Path(temp_dir).rglob('*.mcap'))


def _write_rosbags(
    temp_dir: str,
    msg,
    typestore,
    topic: str = '/rosbags',
    *,
    timestamp: int = 0,
) -> tuple[Path, int]:
    with Writer(Path(temp_dir) / 'rosbags', version=9, storage_plugin=StoragePlugin.MCAP) as writer:
        connection = writer.add_connection(topic, msg.__msgtype__, typestore=typestore)
        writer.write(connection, timestamp, typestore.serialize_cdr(msg, msg.__msgtype__))
    return _find_mcap_file(temp_dir), connection.id


def _make_header(typestore: Typestore, frame_id: str = "frame", sec: int = 1, nanosec: int = 2):
    Header = typestore.types["std_msgs/msg/Header"]
    Time = typestore.types["builtin_interfaces/msg/Time"]
    return Header(stamp=Time(sec=sec, nanosec=nanosec), frame_id=frame_id)


@pytest.fixture(params=[Stores.ROS2_JAZZY, Stores.ROS2_HUMBLE])
def typestore(request):
    return get_typestore(request.param)


def test_vector3_rosbags(typestore: Typestore):
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = Vector3(x=1.0, y=2.0, z=3.0)
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.x == 1.0
        assert messages[0].data.y == 2.0
        assert messages[0].data.z == 3.0
        # TODO: Test the type somehow?


def test_pose_with_covariance_rosbags(typestore: Typestore):
    PoseWithCovariance = typestore.types['geometry_msgs/msg/PoseWithCovariance']
    Pose = typestore.types['geometry_msgs/msg/Pose']
    Point = typestore.types['geometry_msgs/msg/Point']
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']

    msg = PoseWithCovariance(
        pose=Pose(
            position=Point(x=1.0, y=2.0, z=3.0),
            orientation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
        ),
        covariance=np.array([float(i) for i in range(36)]),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.pose.position.x == 1.0
        assert messages[0].data.pose.position.y == 2.0
        assert messages[0].data.pose.position.z == 3.0
        assert messages[0].data.pose.orientation.x == 0.0
        assert messages[0].data.pose.orientation.y == 0.0
        assert messages[0].data.pose.orientation.z == 0.0
        assert messages[0].data.pose.orientation.w == 1.0
        assert messages[0].data.covariance == [float(i) for i in range(36)]


def test_vector3_stamped_rosbags(typestore: Typestore):
    Vector3Stamped = typestore.types['geometry_msgs/msg/Vector3Stamped']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = Vector3Stamped(
        header=_make_header(typestore),
        vector=Vector3(x=1.0, y=2.0, z=3.0)
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.vector.x == 1.0
        assert messages[0].data.vector.y == 2.0
        assert messages[0].data.vector.z == 3.0


def test_pose_rosbags(typestore: Typestore):
    Pose = typestore.types['geometry_msgs/msg/Pose']
    Point = typestore.types['geometry_msgs/msg/Point']
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']

    msg = Pose(
        position=Point(x=1.0, y=2.0, z=3.0),
        orientation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.position.x == 1.0
        assert messages[0].data.position.y == 2.0
        assert messages[0].data.position.z == 3.0
        assert messages[0].data.orientation.x == 0.0
        assert messages[0].data.orientation.y == 0.0
        assert messages[0].data.orientation.z == 0.0
        assert messages[0].data.orientation.w == 1.0


def test_inertia_stamped_rosbags(typestore: Typestore):
    InertiaStamped = typestore.types['geometry_msgs/msg/InertiaStamped']
    Inertia = typestore.types['geometry_msgs/msg/Inertia']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = InertiaStamped(
        header=_make_header(typestore),
        inertia=Inertia(
            m=1.0,
            com=Vector3(x=1.0, y=2.0, z=3.0),
            ixx=0.1,
            ixy=0.2,
            ixz=0.3,
            iyy=0.4,
            iyz=0.5,
            izz=0.6,
        ),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.inertia.m == 1.0
        assert messages[0].data.inertia.com.x == 1.0
        assert messages[0].data.inertia.com.y == 2.0
        assert messages[0].data.inertia.com.z == 3.0
        assert messages[0].data.inertia.ixx == 0.1
        assert messages[0].data.inertia.ixy == 0.2
        assert messages[0].data.inertia.ixz == 0.3
        assert messages[0].data.inertia.iyy == 0.4
        assert messages[0].data.inertia.iyz == 0.5
        assert messages[0].data.inertia.izz == 0.6


def test_transform_stamped_rosbags(typestore: Typestore):
    TransformStamped = typestore.types['geometry_msgs/msg/TransformStamped']
    Transform = typestore.types['geometry_msgs/msg/Transform']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']

    msg = TransformStamped(
        header=_make_header(typestore),
        child_frame_id='child',
        transform=Transform(
            translation=Vector3(x=1.0, y=2.0, z=3.0),
            rotation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
        ),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.child_frame_id == 'child'
        assert messages[0].data.transform.translation.x == 1.0
        assert messages[0].data.transform.translation.y == 2.0
        assert messages[0].data.transform.translation.z == 3.0
        assert messages[0].data.transform.rotation.x == 0.0
        assert messages[0].data.transform.rotation.y == 0.0
        assert messages[0].data.transform.rotation.z == 0.0
        assert messages[0].data.transform.rotation.w == 1.0


def test_twist_rosbags(typestore: Typestore):
    Twist = typestore.types['geometry_msgs/msg/Twist']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = Twist(
        linear=Vector3(x=1.0, y=2.0, z=3.0),
        angular=Vector3(x=4.0, y=5.0, z=6.0),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.linear.x == 1.0
        assert messages[0].data.linear.y == 2.0
        assert messages[0].data.linear.z == 3.0
        assert messages[0].data.angular.x == 4.0
        assert messages[0].data.angular.y == 5.0
        assert messages[0].data.angular.z == 6.0


def test_accel_with_covariance_rosbags(typestore: Typestore):
    AccelWithCovariance = typestore.types['geometry_msgs/msg/AccelWithCovariance']
    Accel = typestore.types['geometry_msgs/msg/Accel']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = AccelWithCovariance(
        accel=Accel(
            linear=Vector3(x=1.0, y=2.0, z=3.0),
            angular=Vector3(x=4.0, y=5.0, z=6.0),
        ),
        covariance=np.array([float(i) for i in range(36)]),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.accel.linear.x == 1.0
        assert messages[0].data.accel.linear.y == 2.0
        assert messages[0].data.accel.linear.z == 3.0
        assert messages[0].data.accel.angular.x == 4.0
        assert messages[0].data.accel.angular.y == 5.0
        assert messages[0].data.accel.angular.z == 6.0
        assert messages[0].data.covariance == [float(i) for i in range(36)]


def test_quaternion_stamped_rosbags(typestore: Typestore):
    QuaternionStamped = typestore.types['geometry_msgs/msg/QuaternionStamped']
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']

    msg = QuaternionStamped(
        header=_make_header(typestore),
        quaternion=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.quaternion.x == 0.0
        assert messages[0].data.quaternion.y == 0.0
        assert messages[0].data.quaternion.z == 0.0
        assert messages[0].data.quaternion.w == 1.0


def test_transform_rosbags(typestore: Typestore):
    Transform = typestore.types['geometry_msgs/msg/Transform']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']

    msg = Transform(
        translation=Vector3(x=1.0, y=2.0, z=3.0),
        rotation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.translation.x == 1.0
        assert messages[0].data.translation.y == 2.0
        assert messages[0].data.translation.z == 3.0
        assert messages[0].data.rotation.x == 0.0
        assert messages[0].data.rotation.y == 0.0
        assert messages[0].data.rotation.z == 0.0
        assert messages[0].data.rotation.w == 1.0


def test_wrench_stamped_rosbags(typestore: Typestore):
    WrenchStamped = typestore.types['geometry_msgs/msg/WrenchStamped']
    Wrench = typestore.types['geometry_msgs/msg/Wrench']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = WrenchStamped(
        header=_make_header(typestore),
        wrench=Wrench(
            force=Vector3(x=1.0, y=2.0, z=3.0),
            torque=Vector3(x=4.0, y=5.0, z=6.0),
        ),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.wrench.force.x == 1.0
        assert messages[0].data.wrench.force.y == 2.0
        assert messages[0].data.wrench.force.z == 3.0
        assert messages[0].data.wrench.torque.x == 4.0
        assert messages[0].data.wrench.torque.y == 5.0
        assert messages[0].data.wrench.torque.z == 6.0


def test_accel_rosbags(typestore: Typestore):
    Accel = typestore.types['geometry_msgs/msg/Accel']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = Accel(
        linear=Vector3(x=1.0, y=2.0, z=3.0),
        angular=Vector3(x=4.0, y=5.0, z=6.0),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.linear.x == 1.0
        assert messages[0].data.linear.y == 2.0
        assert messages[0].data.linear.z == 3.0
        assert messages[0].data.angular.x == 4.0
        assert messages[0].data.angular.y == 5.0
        assert messages[0].data.angular.z == 6.0


def test_polygon_rosbags(typestore: Typestore):
    Polygon = typestore.types['geometry_msgs/msg/Polygon']
    Point32 = typestore.types['geometry_msgs/msg/Point32']

    msg = Polygon(
        points=[
            Point32(x=1.0, y=2.0, z=3.0),
            Point32(x=4.0, y=5.0, z=6.0),
        ]
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert len(messages[0].data.points) == 2
        assert messages[0].data.points[0].x == 1.0
        assert messages[0].data.points[0].y == 2.0
        assert messages[0].data.points[0].z == 3.0
        assert messages[0].data.points[1].x == 4.0
        assert messages[0].data.points[1].y == 5.0
        assert messages[0].data.points[1].z == 6.0


def test_pose2d_rosbags(typestore: Typestore):
    Pose2D = typestore.types['geometry_msgs/msg/Pose2D']

    msg = Pose2D(x=1.0, y=2.0, theta=3.0)
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.x == 1.0
        assert messages[0].data.y == 2.0
        assert messages[0].data.theta == 3.0


def test_pose_stamped_rosbags(typestore: Typestore):
    PoseStamped = typestore.types['geometry_msgs/msg/PoseStamped']
    Pose = typestore.types['geometry_msgs/msg/Pose']
    Point = typestore.types['geometry_msgs/msg/Point']
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']

    msg = PoseStamped(
        header=_make_header(typestore),
        pose=Pose(
            position=Point(x=1.0, y=2.0, z=3.0),
            orientation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
        ),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.pose.position.x == 1.0
        assert messages[0].data.pose.position.y == 2.0
        assert messages[0].data.pose.position.z == 3.0
        assert messages[0].data.pose.orientation.x == 0.0
        assert messages[0].data.pose.orientation.y == 0.0
        assert messages[0].data.pose.orientation.z == 0.0
        assert messages[0].data.pose.orientation.w == 1.0


def test_accel_stamped_rosbags(typestore: Typestore):
    AccelStamped = typestore.types['geometry_msgs/msg/AccelStamped']
    Accel = typestore.types['geometry_msgs/msg/Accel']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = AccelStamped(
        header=_make_header(typestore),
        accel=Accel(
            linear=Vector3(x=1.0, y=2.0, z=3.0),
            angular=Vector3(x=4.0, y=5.0, z=6.0),
        ),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.accel.linear.x == 1.0
        assert messages[0].data.accel.linear.y == 2.0
        assert messages[0].data.accel.linear.z == 3.0
        assert messages[0].data.accel.angular.x == 4.0
        assert messages[0].data.accel.angular.y == 5.0
        assert messages[0].data.accel.angular.z == 6.0


def test_accel_with_covariance_stamped_rosbags(typestore: Typestore):
    AccelWithCovarianceStamped = typestore.types['geometry_msgs/msg/AccelWithCovarianceStamped']
    AccelWithCovariance = typestore.types['geometry_msgs/msg/AccelWithCovariance']
    Accel = typestore.types['geometry_msgs/msg/Accel']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = AccelWithCovarianceStamped(
        header=_make_header(typestore),
        accel=AccelWithCovariance(
            accel=Accel(
                linear=Vector3(x=1.0, y=2.0, z=3.0),
                angular=Vector3(x=4.0, y=5.0, z=6.0),
            ),
            covariance=np.array([float(i) for i in range(36)]),
        ),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.accel.accel.linear.x == 1.0
        assert messages[0].data.accel.accel.linear.y == 2.0
        assert messages[0].data.accel.accel.linear.z == 3.0
        assert messages[0].data.accel.accel.angular.x == 4.0
        assert messages[0].data.accel.accel.angular.y == 5.0
        assert messages[0].data.accel.accel.angular.z == 6.0
        assert messages[0].data.accel.covariance == [float(i) for i in range(36)]


def test_quaternion_rosbags(typestore: Typestore):
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']

    msg = Quaternion(x=0.0, y=0.0, z=0.0, w=1.0)
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.x == 0.0
        assert messages[0].data.y == 0.0
        assert messages[0].data.z == 0.0
        assert messages[0].data.w == 1.0


def test_point_rosbags(typestore: Typestore):
    Point = typestore.types['geometry_msgs/msg/Point']

    msg = Point(x=1.0, y=2.0, z=3.0)
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.x == 1.0
        assert messages[0].data.y == 2.0
        assert messages[0].data.z == 3.0


def test_inertia_rosbags(typestore: Typestore):
    Inertia = typestore.types['geometry_msgs/msg/Inertia']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = Inertia(
        m=1.0,
        com=Vector3(x=1.0, y=2.0, z=3.0),
        ixx=0.1,
        ixy=0.2,
        ixz=0.3,
        iyy=0.4,
        iyz=0.5,
        izz=0.6,
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.m == 1.0
        assert messages[0].data.com.x == 1.0
        assert messages[0].data.com.y == 2.0
        assert messages[0].data.com.z == 3.0
        assert messages[0].data.ixx == 0.1
        assert messages[0].data.ixy == 0.2
        assert messages[0].data.ixz == 0.3
        assert messages[0].data.iyy == 0.4
        assert messages[0].data.iyz == 0.5
        assert messages[0].data.izz == 0.6


def test_polygon_stamped_rosbags(typestore: Typestore):
    PolygonStamped = typestore.types['geometry_msgs/msg/PolygonStamped']
    Polygon = typestore.types['geometry_msgs/msg/Polygon']
    Point32 = typestore.types['geometry_msgs/msg/Point32']

    msg = PolygonStamped(
        header=_make_header(typestore),
        polygon=Polygon(
            points=[
                Point32(x=1.0, y=2.0, z=3.0),
                Point32(x=4.0, y=5.0, z=6.0),
            ]
        ),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert len(messages[0].data.polygon.points) == 2
        assert messages[0].data.polygon.points[0].x == 1.0
        assert messages[0].data.polygon.points[0].y == 2.0
        assert messages[0].data.polygon.points[0].z == 3.0
        assert messages[0].data.polygon.points[1].x == 4.0
        assert messages[0].data.polygon.points[1].y == 5.0
        assert messages[0].data.polygon.points[1].z == 6.0


def test_pose_array_rosbags(typestore: Typestore):
    PoseArray = typestore.types['geometry_msgs/msg/PoseArray']
    Pose = typestore.types['geometry_msgs/msg/Pose']
    Point = typestore.types['geometry_msgs/msg/Point']
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']

    msg = PoseArray(
        header=_make_header(typestore),
        poses=[
            Pose(
                position=Point(x=1.0, y=2.0, z=3.0),
                orientation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
            ),
            Pose(
                position=Point(x=4.0, y=5.0, z=6.0),
                orientation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
            ),
        ],
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert len(messages[0].data.poses) == 2
        assert messages[0].data.poses[0].position.x == 1.0
        assert messages[0].data.poses[0].position.y == 2.0
        assert messages[0].data.poses[0].position.z == 3.0
        assert messages[0].data.poses[0].orientation.x == 0.0
        assert messages[0].data.poses[0].orientation.y == 0.0
        assert messages[0].data.poses[0].orientation.z == 0.0
        assert messages[0].data.poses[0].orientation.w == 1.0
        assert messages[0].data.poses[1].position.x == 4.0
        assert messages[0].data.poses[1].position.y == 5.0
        assert messages[0].data.poses[1].position.z == 6.0
        assert messages[0].data.poses[1].orientation.x == 0.0
        assert messages[0].data.poses[1].orientation.y == 0.0
        assert messages[0].data.poses[1].orientation.z == 0.0
        assert messages[0].data.poses[1].orientation.w == 1.0


def test_point_stamped_rosbags(typestore: Typestore):
    PointStamped = typestore.types['geometry_msgs/msg/PointStamped']
    Point = typestore.types['geometry_msgs/msg/Point']

    msg = PointStamped(
        header=_make_header(typestore),
        point=Point(x=1.0, y=2.0, z=3.0)
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.point.x == 1.0
        assert messages[0].data.point.y == 2.0
        assert messages[0].data.point.z == 3.0


def test_point32_rosbags(typestore: Typestore):
    Point32 = typestore.types['geometry_msgs/msg/Point32']

    msg = Point32(x=1.0, y=2.0, z=3.0)
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.x == 1.0
        assert messages[0].data.y == 2.0
        assert messages[0].data.z == 3.0


def test_wrench_rosbags(typestore: Typestore):
    Wrench = typestore.types['geometry_msgs/msg/Wrench']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = Wrench(
        force=Vector3(x=1.0, y=2.0, z=3.0),
        torque=Vector3(x=4.0, y=5.0, z=6.0),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.force.x == 1.0
        assert messages[0].data.force.y == 2.0
        assert messages[0].data.force.z == 3.0
        assert messages[0].data.torque.x == 4.0
        assert messages[0].data.torque.y == 5.0
        assert messages[0].data.torque.z == 6.0


def test_twist_stamped_rosbags(typestore: Typestore):
    TwistStamped = typestore.types['geometry_msgs/msg/TwistStamped']
    Twist = typestore.types['geometry_msgs/msg/Twist']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = TwistStamped(
        header=_make_header(typestore),
        twist=Twist(
            linear=Vector3(x=1.0, y=2.0, z=3.0),
            angular=Vector3(x=4.0, y=5.0, z=6.0),
        ),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.twist.linear.x == 1.0
        assert messages[0].data.twist.linear.y == 2.0
        assert messages[0].data.twist.linear.z == 3.0
        assert messages[0].data.twist.angular.x == 4.0
        assert messages[0].data.twist.angular.y == 5.0
        assert messages[0].data.twist.angular.z == 6.0


def test_twist_with_covariance_stamped_rosbags(typestore: Typestore):
    TwistWithCovarianceStamped = typestore.types['geometry_msgs/msg/TwistWithCovarianceStamped']
    TwistWithCovariance = typestore.types['geometry_msgs/msg/TwistWithCovariance']
    Twist = typestore.types['geometry_msgs/msg/Twist']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = TwistWithCovarianceStamped(
        header=_make_header(typestore),
        twist=TwistWithCovariance(
            twist=Twist(
                linear=Vector3(x=1.0, y=2.0, z=3.0),
                angular=Vector3(x=4.0, y=5.0, z=6.0),
            ),
            covariance=np.array([float(i) for i in range(36)]),
        ),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.twist.twist.linear.x == 1.0
        assert messages[0].data.twist.twist.linear.y == 2.0
        assert messages[0].data.twist.twist.linear.z == 3.0
        assert messages[0].data.twist.twist.angular.x == 4.0
        assert messages[0].data.twist.twist.angular.y == 5.0
        assert messages[0].data.twist.twist.angular.z == 6.0
        assert messages[0].data.twist.covariance == [float(i) for i in range(36)]


def test_twist_with_covariance_rosbags(typestore: Typestore):
    TwistWithCovariance = typestore.types['geometry_msgs/msg/TwistWithCovariance']
    Twist = typestore.types['geometry_msgs/msg/Twist']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = TwistWithCovariance(
        twist=Twist(
            linear=Vector3(x=1.0, y=2.0, z=3.0),
            angular=Vector3(x=4.0, y=5.0, z=6.0),
        ),
        covariance=np.array([float(i) for i in range(36)]),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.twist.linear.x == 1.0
        assert messages[0].data.twist.linear.y == 2.0
        assert messages[0].data.twist.linear.z == 3.0
        assert messages[0].data.twist.angular.x == 4.0
        assert messages[0].data.twist.angular.y == 5.0
        assert messages[0].data.twist.angular.z == 6.0
        assert messages[0].data.covariance == [float(i) for i in range(36)]


def test_pose_with_covariance_stamped_rosbags(typestore: Typestore):
    PoseWithCovarianceStamped = typestore.types['geometry_msgs/msg/PoseWithCovarianceStamped']
    PoseWithCovariance = typestore.types['geometry_msgs/msg/PoseWithCovariance']
    Pose = typestore.types['geometry_msgs/msg/Pose']
    Point = typestore.types['geometry_msgs/msg/Point']
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']

    msg = PoseWithCovarianceStamped(
        header=_make_header(typestore),
        pose=PoseWithCovariance(
            pose=Pose(
                position=Point(x=1.0, y=2.0, z=3.0),
                orientation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
            ),
            covariance=np.array([float(i) for i in range(36)]),
        ),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.pose.pose.position.x == 1.0
        assert messages[0].data.pose.pose.position.y == 2.0
        assert messages[0].data.pose.pose.position.z == 3.0
        assert messages[0].data.pose.pose.orientation.x == 0.0
        assert messages[0].data.pose.pose.orientation.y == 0.0
        assert messages[0].data.pose.pose.orientation.z == 0.0
        assert messages[0].data.pose.pose.orientation.w == 1.0
        assert messages[0].data.pose.covariance == [float(i) for i in range(36)]
